// 
//  fft.cpp
//  
//  
//  Created by Ben Swieskowski
// 
//  Third party code used:
//  - chuck_fft.h/cpp based on the FFT implementation in CARL
//  - Thread.h/cpp from STK 


#ifdef __MACOSX_CORE__
  #include <GLUT/glut.h>
#else
  #include <GL/gl.h>
  #include <GL/glu.h>
  #include <GL/glut.h>
#endif


#include <iostream>
#include <math.h>
#include "RtAudio.h"
#include <sstream>

#include "chuck_fft.h"
#include "Thread.h"


#define MY_SRATE 44100
#define SAMPLE double 
#define ZPF 1

using namespace std;

// ===========
// = Globals =
// ===========
GLsizei g_width = 800; 
GLsizei g_height = 600;

Mutex g_mutex;

float * g_fftBuff;
float * g_window;
unsigned int g_buffSize = 1024;
// next 3 vars only hold checksums, not entire spectra
float g_previous_sum = 1.0;
float g_last_200[200];
float g_max_current = .1;
// tuning params
int g_speed = 7;
// this array holds 200 previous spectra
complex g_ffts[200][512];

float beats[100]; // positions of beats as they move across screen

float notes[100]; // heights of notes above each beat

int hits[100]; // keeps track of whether notes have been hit

int linedup[100];

int score = 0;

int beatsLeft = 100;

float g_paddleHeight = .5;
int g_previousMaxIndex = 0;
int g_previousStatus = 0; // 0 = not whistling, 1 = whistling

double burstX[100];
double burstY[100];
double burstXV[100];
double burstYV[100];
double burstXO[100];
double burstYO[100];
int burst = -1;
int burstFrame = 0;








//-----------------------------------------------------------------------------
// Defines a point in a 3D space (coords x, y and z)
//-----------------------------------------------------------------------------




struct pt3d
{
    pt3d( GLfloat x, GLfloat y, GLfloat z ) : x(x), y(y), z(z) {};
    
    float x;
    float y;
    float z;
};



// =======================
// = Function prototypes =
// =======================
void idleFunc( );
void displayFunc( );
void reshapeFunc( int width, int height );
void keyboardFunc( unsigned char, int, int );
void mouseFunc( int button, int state, int x, int y );
void specialFunc( int key, int x, int y );
void initialize( );
void changeLookAt( pt3d look_from, pt3d look_to, pt3d head_up );
void drawAxis();
void drawNote(double x, double y);
void detectCollision();
void drawBeats();
void drawNotes();
void initBurst();
void drawBurst();
void drawPaddle();
void drawText();
void drawMeasures();
void detectWhistle();



// ==========================
// = Cammera placement vars =
// ==========================
// Camera control global variables
pt3d g_look_from( 0, 0, 1);
pt3d g_look_to( 0, 0, 0 );
pt3d g_head_up( 0, 1, 0 );

// ============
// = GL stuff =
// ============
//-----------------------------------------------------------------------------
// Name: initialize( )
// Desc: sets initial OpenGL states
//       also initializes any application data
//-----------------------------------------------------------------------------
void initialize()
{

    // =================
    // = OpenGL & GLUT =
    // =================

    // set the GL clear color - use when the color buffer is cleared
    glClearColor( 0.0f, 0.0f, 0.0f, 1.0f );
        
    // set the shading model to 'smooth'
    glShadeModel( GL_SMOOTH );
    // enable depth
    glEnable( GL_DEPTH_TEST );
    // set the front faces of polygons
    glFrontFace( GL_CCW );
    // set fill mode
    glPolygonMode( GL_FRONT_AND_BACK, GL_FILL );

    // Enable transparency
    glEnable(GL_BLEND);
    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
          
    // seed random number generator
    srand( time(NULL) );
}


//-----------------------------------------------------------------------------
// Name: reshapeFunc( )
// Desc: called when window size changes
//-----------------------------------------------------------------------------
void reshapeFunc( int w, int h )
{
    // save the new window size
    g_width = (GLsizei)w; 
    g_height = (GLsizei)h;
    // map the view port to the client area
    glViewport( 0, 0, (GLsizei)w, (GLsizei)h );
    // set the matrix mode to project
    glMatrixMode( GL_PROJECTION );
    // load the identity matrix
    glLoadIdentity( );
    // create the viewing frustum
    gluPerspective( 45.0, (GLfloat) w / (GLfloat) h, 0.1, 300.0 );
    // set the matrix mode to modelview
    glMatrixMode( GL_MODELVIEW );
    // load the identity matrix
    glLoadIdentity( );
    // position the view point
    changeLookAt( g_look_from, g_look_to, g_head_up );
}


//-----------------------------------------------------------------------------
// Name: keyboardFunc( )
// Desc: key event
//-----------------------------------------------------------------------------
void keyboardFunc( unsigned char key, int x, int y )
{
    switch( key )
    { 
	case '<':
	case ',':
		g_paddleHeight -= .05;
		break;
	case '>':
	case '.':
		g_paddleHeight += .05;
		break;
	case 'Q':
	case 'q':
	    cerr << "Bye!" << endl;
	    exit(1);
	    break;
    }
    glutPostRedisplay( );
}

//-----------------------------------------------------------------------------
// Name: idleFunc( )
// Desc: callback from GLUT
//-----------------------------------------------------------------------------
void idleFunc( )
{
    // render the scene
    glutPostRedisplay( );
}


//-----------------------------------------------------------------------------
// Name: displayFunc( )
// Desc: callback function invoked to draw the client area
//-----------------------------------------------------------------------------
void displayFunc( )
{
	
	// this loop does nothing but control framerate
	for (int i = 0; i < 10 - g_speed; i++){
		for (int j = 0; j < 1000000; j++)
		{
			int x = 5+2;
		}
	}
	
    // clear the color and depth buffers
    glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );

    // save the current transformation
    glPushMatrix();
        
        // set the view point (camera)
        changeLookAt( g_look_from, g_look_to, g_head_up );

        glLineWidth( 1.0f );  
  
        glTranslatef( -0.5, -0.5, 0 );
   
		detectWhistle();
			
		drawPaddle();
				
		drawBeats();
				
		drawNotes();	
				
		drawBurst();			
				
		drawMeasures();
				
		detectCollision();
		
		drawText();

		// move beats
		for (int i = 0; i < 100; i++)
			beats[i]-= .001;												
	  
    // restore state
    glPopMatrix();

    // flush!
    glFlush( );
    // swap the double buffer
    glutSwapBuffers( );
}

void drawMeasures(){
	glColor4f(1, 1, 1, 1);
	for(int i = 0; i < 25; i++)
	{
		glBegin(GL_LINE_STRIP);
			glVertex3f(beats[i*4] + .115 - .23, -.25,0);
			glVertex3f(beats[i*4] + .115 - .23, 1.25,0);					
		glEnd();
	}
}

void detectWhistle(){
	g_mutex.lock();
	complex * fft = (complex *)g_fftBuff;
	g_mutex.unlock();
	
	// copy new spectrum into spectra array, and compute checksum
	float sum1 = 0;
	for (int i = 0; i < g_buffSize/2 * ZPF; i ++)
	{
		g_ffts[0][i] = fft[i];
		sum1 += cmp_abs(fft[i]);
	}
	
	// BEGIN DETECT WHISTLE ALGO
	
	// find max height and location of max height
	float thisMax2 = .0;
	int thisMaxIndex2 = 0;
	for (int i = 0; i < g_buffSize/2 * ZPF; i++)
	{
		// find frequency with maximum amplitude 
		if ( cmp_abs(fft[i]) > thisMax2 ){
			thisMax2 = cmp_abs(fft[i]);
			thisMaxIndex2 = i;
		}	
	}
	
	float middle5sum = 0;
	for (int i = 0; i < 5; i ++){
		middle5sum += cmp_abs(fft[thisMaxIndex2 - 2 + i]);
	}
	
	// used to be middle5sum * 15 or * 8
	if (middle5sum * 15 > sum1 && thisMax2 > .01 && thisMaxIndex2 > 5 && thisMaxIndex2 < 200){
		
		if (abs(g_previousMaxIndex - thisMaxIndex2) < 2 && g_previousStatus > 1)
			g_paddleHeight += ( thisMaxIndex2 - g_previousMaxIndex ) * .03;
		g_previousStatus++;
		
	}
	else {
		g_previousStatus = 0;
	}
	
	// END DETECT WHISTLE ALGO
	
	g_previousMaxIndex = thisMaxIndex2;
	
	
	// make sure we aren't saving same buffer twice in a row
	if (g_previous_sum != sum1)
	{		
		// advance array of spectra
		for (int j = 198; j >=0; j--){
			for (int h = 0; h < g_buffSize * ZPF / 2; h++)
				g_ffts[j+1][h] = g_ffts[j][h];
		}	
	}
	
	// save checksum
	g_previous_sum = sum1;
	
	// constrain paddle vertically
	if (g_paddleHeight < 0)
		g_paddleHeight = 0;
	
	if (g_paddleHeight > 1)
		g_paddleHeight = 1;
	
}

void drawText(){
	
	std::string score2;
	std::stringstream out;
	out << score;
	score2 = out.str();
	
	std::string beatsLeft2;
	std::stringstream out2;
	out2 << beatsLeft;
	beatsLeft2 = out2.str();
	
	glRasterPos3d(0,1.28,0);

	string s = "Score: " + score2;
	void * font = GLUT_BITMAP_9_BY_15;
	for (string::iterator i = s.begin(); i != s.end(); ++i)
	{
	    char c = *i;
	    glutBitmapCharacter(font, c);
	}
	
	glRasterPos3d(-.5,1.28,0);

	string s2 = "Notes Left: " + beatsLeft2;
	for (string::iterator i = s2.begin(); i != s2.end(); ++i)
	{
	    char c = *i;
	    glutBitmapCharacter(font, c);
	}
	
}

void drawPaddle(){
	glColor4f(0,0,1,1);
	
	glBegin(GL_TRIANGLES);					

				glVertex3f(-.5, g_paddleHeight, 0);
				glVertex3f(-.3, g_paddleHeight + .025, 0);
				glVertex3f(-.5, g_paddleHeight + .05, 0);
				glVertex3f(-.5, g_paddleHeight, 0);
		
    glEnd();
	
}

void drawNotes(){
	
	for(int i = 0; i < 100; i++){	
		if (hits[i] == 1) // note has been hit, set color to green
			glColor4f(0, 1, 0, 1);
		else if (hits[i] == -1) // note has been missed set color to red
			glColor4f(1, 0, 0, 1);
		else { // note not yet played
			if (linedup[i] == 0) // not in line with pointer, set color to white
				glColor4f(1, 1, 1, 1);
			else { // in line with pointer, make flashing green
				double tempCol;
				tempCol = (int(500 * (beats[i] + 20)) % 15 ) / 15.0;
				glColor4f(tempCol,1, tempCol,1);
			}
		}
		drawNote(beats[i], notes[i]);
	}	
	
}

void drawBeats(){
	glColor4f(1,1,1,1);
	for(int i = 0; i < 100; i++){
		glBegin(GL_LINE_STRIP);					
        	glVertex3f(beats[i] - .03,-.23,0);
          	glVertex3f(beats[i] + .03 - .03,-.2,0);							              			                    				
    	glEnd();
	}
	for(int i = 0; i < 100; i++){
		glBegin(GL_LINE_STRIP);					
        	glVertex3f(beats[i] - .03,-.2,0);
			glVertex3f(beats[i] + .03 - .03,-.23,0);                   	
    	glEnd();
	}	
}

void drawBurst() {
	if (burst != - 1) {
		glColor4f(1, 1.0 - double(burstFrame)/100, 0, 1.0 - double(burstFrame)/200 );
		
		burstFrame++;
		for (int i = 0; i < 100; i++){
			glBegin(GL_LINE_STRIP);
				glVertex3f(burstX[i], burstY[i], 0);	
				glVertex3f(burstX[i] + burstXO[i], burstY[i] + burstYO[i], 0);	
				
			glEnd();
			
			burstX[i] += burstXV[i]/100.0;
			burstY[i] += burstYV[i]/100.0;
		}
	}
}

void initBurst(){
	for (int i = 0; i < 100; i++){
		burstX[i] = -.3;
		burstY[i] = g_paddleHeight + .025;
		double tempAngle = 2*3.14 * (float)rand()/(float)RAND_MAX;
		double tempVel = (float)rand()/(float)RAND_MAX + .2;
		burstXV[i] = cos(tempAngle) * tempVel;
		burstYV[i] = sin(tempAngle) * tempVel;
		burstXO[i] = ((float)rand()/(float)RAND_MAX - .5) / 10.0;
		burstYO[i] = ((float)rand()/(float)RAND_MAX - .5) / 10.0;
		burstFrame = 0;	
	}
	
}

void detectCollision(){
	
	for (int i = 0; i < 100; i ++){
		float dist = fabs(beats[i] + .3);
		if (dist < .000499){ // if x coords are aligned
			
			beatsLeft --;
			
			if (g_paddleHeight + .025 > notes[i] && g_paddleHeight + .025 < notes[i] + .12){
				//if (g_paddleHeight < notes[i] + .12 )
				//{
					score++;
					hits[i] = 1;
					burst = 1;
					initBurst();
				//}
			}
			else {
				hits[i] = -1;
			}
			
		}
		if (beats[i] < .31 && hits[i] == 0)
			hits[i] == -1;
	}
	
	for (int i = 0; i < 100; i ++){
		
		if (g_paddleHeight + .025 > notes[i] && g_paddleHeight + .025 < notes[i] + .12){
			//if (g_paddleHeight < notes[i] + .12 )
			//{
			linedup[i] = 1;
			//}
		}
		else {
			linedup[i] = 0;
		}
	
	}
	
}

void drawNote( double xloc, double yloc){
	double x1, x2, y1, y2;
	
	
	glBegin(GL_LINE_STRIP);					
    
		glVertex3f(xloc, yloc + .12, 0);
		glVertex3f(xloc, yloc + .025, 0);
    glEnd();

	double x = xloc - .025; 
	double y = yloc + .025;
	x1 = x;
	y1 = y;
	float radius = 0.025;

	glBegin(GL_TRIANGLES);  
		for(int i=0;i<=360;i++)
		{
			double angle=(float)(((double)i)/57.29577957795135);   
		    double x2=x+(radius*(float)sin((double)angle));
		    double y2=y+(radius*(float)cos((double)angle));             
		    glVertex3d(x,y,0);
		    glVertex3d(x1,y1,0);
		    glVertex3d(x2,y2,0);
		    y1=y2;
		    x1=x2;
	 	}
	glEnd();
}

//-----------------------------------------------------------------------------
// name: changeLookAt()
// desc: changes the point of view
//-----------------------------------------------------------------------------
void changeLookAt( pt3d look_from, pt3d look_to, pt3d head_up )
{
    gluLookAt(  look_from.x, look_from.y, look_from.z, 
                look_to.x, look_to.y, look_to.z, 
                head_up.x, head_up.y, head_up.z );
}



// =========
// = Audio =
// =========
// Audio callback
int audioCallback( void * outputBuffer, void * inputBuffer, 
            unsigned int bufferSize, double streamTime,
            RtAudioStreamStatus status, void * userData )
{
    SAMPLE * out = (SAMPLE *)outputBuffer;
    SAMPLE * in = (SAMPLE *)inputBuffer;
        
    g_mutex.lock();
        
    // zero out the fft buffer (required only if ZPF > 1)
    //memset( g_fftBuff, 0, ZPF * bufferSize * sizeof(float) );
    
    for(size_t i = 0; i < bufferSize; ++i)
    {
        out[i] = 0;
        // "static" sine wave for testing
        //g_fftBuff[i] = ::sin( 2 * M_PI * 2200.0f * i / MY_SRATE );
        g_fftBuff[i] = in[i];
    }
    
    // apply window to the buffer of audio
    

    // compute the fft
    rfft( g_fftBuff, ZPF * bufferSize / 2, FFT_FORWARD );
    
    g_mutex.unlock();
    
    return 0;
}


// ========
// = Main =
// ========
// Entry point
int main (int argc, char *argv[])
{
	// seed rands
	srand((unsigned)time(0));
	
    // initialize last 1000 checksums used for tuning color sensitivty
	for (int i = 0; i < 200; i++){
		g_last_200[i] = 0;
	}
	
	// initialize hits
	for (int i = 0; i < 100; i++){
		hits[i] = 0;
		linedup[i] = 0;
	}
	
	// initialize beats
	for (int i = 0; i < 100; i++)
		beats[i] = i*.2 - .1;
		
	// initialize notes
	for (int i = 0; i < 100; i++)
	{
		float r = (float)rand()/(float)RAND_MAX;	
		notes[i] = r;
	}
	
    // RtAudio config + init

    // pointer to RtAudio object
    RtAudio *  audio = NULL;

    // create the object
    try
    {
        audio = new RtAudio();
    }
        catch( RtError & err ) {
        err.printMessage();
        exit(1);
    }

    if( audio->getDeviceCount() < 1 )
    {
        // nopes
        cout << "no audio devices found!" << endl;
        exit( 1 );
    }
        
    // let RtAudio print messages to stderr.
    audio->showWarnings( true );

    // set input and output parameters
    RtAudio::StreamParameters iParams, oParams;
    iParams.deviceId = audio->getDefaultInputDevice();
    iParams.nChannels = 1;
    iParams.firstChannel = 0;
    oParams.deviceId = audio->getDefaultOutputDevice();
    oParams.nChannels = 1;
    oParams.firstChannel = 0;
        
    // create stream options
    RtAudio::StreamOptions options;

    // set the callback and start stream
    try
    {
        audio->openStream( &oParams, &iParams, RTAUDIO_FLOAT64, MY_SRATE, &g_buffSize, &audioCallback, NULL, &options);
        
        cerr << "Buffer size defined by RtAudio: " << g_buffSize << endl;
        
        // allocate the buffer for the fft
        g_fftBuff = new float[g_buffSize * ZPF];
        if ( g_fftBuff == NULL ) {
            cerr << "Something went wrong when creating the fft buffers" << endl;
            exit (1);
        }
        
        // allocate the buffer for the time domain window
        g_window = new float[g_buffSize];
        if ( g_window == NULL ) {
            cerr << "Something went wrong when creating the window" << endl;
            exit (1);
        }

        // create a hanning window
        make_window( g_window, g_buffSize );
        
        // start the audio stream
        audio->startStream();
        
        // test RtAudio functionality for reporting latency.
        cout << "stream latency: " << audio->getStreamLatency() << " frames" << endl;
    }
    catch( RtError & err )
    {
        err.printMessage();
        goto cleanup;
    }


    // ============
    // = GL stuff =
    // ============

    // initialize GLUT
    glutInit( &argc, argv );
    // double buffer, use rgb color, enable depth buffer
    glutInitDisplayMode( GLUT_DOUBLE | GLUT_RGB | GLUT_DEPTH );
    // initialize the window size
    glutInitWindowSize( g_width, g_height );
    // set the window postion
    glutInitWindowPosition( 100, 100 );
    // create the window
    glutCreateWindow( "Hello GL" );
    //glutEnterGameMode();

    // set the idle function - called when idle
    glutIdleFunc( idleFunc );
    // set the display function - called when redrawing
    glutDisplayFunc( displayFunc );
    // set the reshape function - called when client area changes
    glutReshapeFunc( reshapeFunc );
    // set the keyboard function - called on keyboard events
    glutKeyboardFunc( keyboardFunc );


    // do our own initialization
    initialize();

    // let GLUT handle the current thread from here
    glutMainLoop();

        
    // if we get here, stop!
    try
    {
        audio->stopStream();
    }
    catch( RtError & err )
    {
        err.printMessage();
    }

    // Clean up
    cleanup:
    if(audio)
    {
        audio->closeStream();
        delete audio;
    }

    
    return 0;
}